@@ -13,7 +13,7 @@ import (
 	"errors"
 	"io"
 	"os"
-	"sort"
+	"slices"
 	"strings"
 )
 
@@ -31,9 +31,14 @@ var ErrXML = errors.New("etree: invalid XML format")
 var cdataPrefix = []byte("<![CDATA[")
 
 // ReadSettings determine the default behavior of the Document's ReadFrom*
-// methods.
+// functions.
 type ReadSettings struct {
-	// CharsetReader to be passed to standard xml.Decoder. Default: nil.
+	// CharsetReader, if non-nil, defines a function to generate
+	// charset-conversion readers, converting from the provided non-UTF-8
+	// charset into UTF-8. If nil, the ReadFrom* functions will use a
+	// "pass-through" CharsetReader that performs no conversion on the reader's
+	// data regardless of the value of the "charset" encoding string. Default:
+	// nil.
 	CharsetReader func(charset string, input io.Reader) (io.Reader, error)
 
 	// Permissive allows input containing common mistakes such as missing tags
@@ -50,20 +55,33 @@ type ReadSettings struct {
 	// preserve them instead of keeping only one. Default: false.
 	PreserveDuplicateAttrs bool
 
+	// ValidateInput forces all ReadFrom* functions to validate that the
+	// provided input is composed of "well-formed"(*) XML before processing it.
+	// If invalid XML is detected, the ReadFrom* functions return an error.
+	// Because this option requires the input to be processed twice, it incurs a
+	// significant performance penalty. Default: false.
+	//
+	// (*) Note that this definition of "well-formed" is in the context of the
+	// go standard library's encoding/xml package. Go's encoding/xml package
+	// does not, in fact, guarantee well-formed XML as specified by the W3C XML
+	// recommendation. See: https://github.com/golang/go/issues/68299
+	ValidateInput bool
+
 	// Entity to be passed to standard xml.Decoder. Default: nil.
 	Entity map[string]string
+
+	// When Permissive is true, AutoClose indicates a set of elements to
+	// consider closed immediately after they are opened, regardless of
+	// whether an end element is present. Commonly set to xml.HTMLAutoClose.
+	// Default: nil.
+	AutoClose []string
 }
 
-// newReadSettings creates a default ReadSettings record.
-func newReadSettings() ReadSettings {
-	return ReadSettings{
-		CharsetReader: func(label string, input io.Reader) (io.Reader, error) {
-			return input, nil
-		},
-		Permissive:    false,
-		PreserveCData: false,
-		Entity:        nil,
-	}
+// defaultCharsetReader is used by the xml decoder when the ReadSettings
+// CharsetReader value is nil. It behaves as a "pass-through", ignoring
+// the requested charset parameter and skipping conversion altogether.
+func defaultCharsetReader(charset string, input io.Reader) (io.Reader, error) {
+	return input, nil
 }
 
 // dup creates a duplicate of the ReadSettings object.
@@ -82,7 +100,7 @@ func (s *ReadSettings) dup() ReadSettings {
 	}
 }
 
-// WriteSettings determine the behavior of the Document's WriteTo* methods.
+// WriteSettings determine the behavior of the Document's WriteTo* functions.
 type WriteSettings struct {
 	// CanonicalEndTags forces the production of XML end tags, even for
 	// elements that have no child elements. Default: false.
@@ -95,7 +113,8 @@ type WriteSettings struct {
 
 	// CanonicalAttrVal forces the production of XML character references for
 	// attribute value characters &, < and ". If false, XML character
-	// references are also produced for > and '. Default: false.
+	// references are also produced for > and '. Ignored when AttrSingleQuote
+	// is true. Default: false.
 	CanonicalAttrVal bool
 
 	// AttrSingleQuote causes attributes to use single quotes (attr='example')
@@ -103,7 +122,7 @@ type WriteSettings struct {
 	// false.
 	AttrSingleQuote bool
 
-	// UseCRLF causes the document's Indent* methods to use a carriage return
+	// UseCRLF causes the document's Indent* functions to use a carriage return
 	// followed by a linefeed ("\r\n") when outputting a newline. If false,
 	// only a linefeed is used ("\n"). Default: false.
 	//
@@ -111,23 +130,12 @@ type WriteSettings struct {
 	UseCRLF bool
 }
 
-// newWriteSettings creates a default WriteSettings record.
-func newWriteSettings() WriteSettings {
-	return WriteSettings{
-		CanonicalEndTags: false,
-		CanonicalText:    false,
-		CanonicalAttrVal: false,
-		AttrSingleQuote:  false,
-		UseCRLF:          false,
-	}
-}
-
 // dup creates a duplicate of the WriteSettings object.
 func (s *WriteSettings) dup() WriteSettings {
 	return *s
 }
 
-// IndentSettings determine the behavior of the Document's Indent* methods.
+// IndentSettings determine the behavior of the Document's Indent* functions.
 type IndentSettings struct {
 	// Spaces indicates the number of spaces to insert for each level of
 	// indentation. Set to etree.NoIndent to remove all indentation. Ignored
@@ -143,7 +151,7 @@ type IndentSettings struct {
 	// for a newline ("\n"). Default: false.
 	UseCRLF bool
 
-	// PreserveLeafWhitespace causes indent methods to preserve whitespace
+	// PreserveLeafWhitespace causes indent functions to preserve whitespace
 	// within XML elements containing only non-CDATA character data. Default:
 	// false.
 	PreserveLeafWhitespace bool
@@ -185,7 +193,7 @@ func getIndentFunc(s *IndentSettings) indentFunc {
 	}
 }
 
-// Writer is the interface that wraps the Write* methods called by each token
+// Writer is the interface that wraps the Write* functions called by each token
 // type's WriteTo function.
 type Writer interface {
 	io.StringWriter
@@ -250,7 +258,7 @@ const (
 
 // CharData may be used to represent simple text data or a CDATA section
 // within an XML document. The Data property should never be modified
-// directly; use the SetData method instead.
+// directly; use the SetData function instead.
 type CharData struct {
 	Data   string // the simple text or CDATA section content
 	parent *Element
@@ -283,9 +291,7 @@ type ProcInst struct {
 // NewDocument creates an XML document without a root element.
 func NewDocument() *Document {
 	return &Document{
-		Element:       Element{Child: make([]Token, 0)},
-		ReadSettings:  newReadSettings(),
-		WriteSettings: newWriteSettings(),
+		Element: Element{Child: make([]Token, 0)},
 	}
 }
 
@@ -347,6 +353,16 @@ func (d *Document) SetRoot(e *Element) {
 // ReadFrom reads XML from the reader 'r' into this document. The function
 // returns the number of bytes read and any error encountered.
 func (d *Document) ReadFrom(r io.Reader) (n int64, err error) {
+	if d.ReadSettings.ValidateInput {
+		b, err := io.ReadAll(r)
+		if err != nil {
+			return 0, err
+		}
+		if err := validateXML(bytes.NewReader(b), d.ReadSettings); err != nil {
+			return 0, err
+		}
+		r = bytes.NewReader(b)
+	}
 	return d.Element.readFrom(r, d.ReadSettings)
 }
 
@@ -358,22 +374,65 @@ func (d *Document) ReadFromFile(filepath string) error {
 		return err
 	}
 	defer f.Close()
+
 	_, err = d.ReadFrom(f)
 	return err
 }
 
 // ReadFromBytes reads XML from the byte slice 'b' into the this document.
 func (d *Document) ReadFromBytes(b []byte) error {
-	_, err := d.ReadFrom(bytes.NewReader(b))
+	if d.ReadSettings.ValidateInput {
+		if err := validateXML(bytes.NewReader(b), d.ReadSettings); err != nil {
+			return err
+		}
+	}
+	_, err := d.Element.readFrom(bytes.NewReader(b), d.ReadSettings)
 	return err
 }
 
 // ReadFromString reads XML from the string 's' into this document.
 func (d *Document) ReadFromString(s string) error {
-	_, err := d.ReadFrom(strings.NewReader(s))
+	if d.ReadSettings.ValidateInput {
+		if err := validateXML(strings.NewReader(s), d.ReadSettings); err != nil {
+			return err
+		}
+	}
+	_, err := d.Element.readFrom(strings.NewReader(s), d.ReadSettings)
 	return err
 }
 
+// validateXML determines if the data read from the reader 'r' contains
+// well-formed XML according to the rules set by the go xml package.
+func validateXML(r io.Reader, settings ReadSettings) error {
+	dec := newDecoder(r, settings)
+	err := dec.Decode(new(interface{}))
+	if err != nil {
+		return err
+	}
+
+	// If there are any trailing tokens after unmarshalling with Decode(),
+	// then the XML input didn't terminate properly.
+	_, err = dec.Token()
+	if err == io.EOF {
+		return nil
+	}
+	return ErrXML
+}
+
+// newDecoder creates an XML decoder for the reader 'r' configured using
+// the provided read settings.
+func newDecoder(r io.Reader, settings ReadSettings) *xml.Decoder {
+	d := xml.NewDecoder(r)
+	d.CharsetReader = settings.CharsetReader
+	if d.CharsetReader == nil {
+		d.CharsetReader = defaultCharsetReader
+	}
+	d.Strict = !settings.Permissive
+	d.Entity = settings.Entity
+	d.AutoClose = settings.AutoClose
+	return d
+}
+
 // WriteTo serializes the document out to the writer 'w'. The function returns
 // the number of bytes written and any error encountered.
 func (d *Document) WriteTo(w io.Writer) (n int64, err error) {
@@ -693,13 +752,34 @@ func (e *Element) findTermCharDataIndex(start int) int {
 }
 
 // CreateElement creates a new element with the specified tag (i.e., name) and
-// adds it as the last child token of this element. The tag may include a
-// prefix followed by a colon.
+// adds it as the last child of element 'e'. The tag may include a prefix
+// followed by a colon.
 func (e *Element) CreateElement(tag string) *Element {
 	space, stag := spaceDecompose(tag)
 	return newElement(space, stag, e)
 }
 
+// CreateChild performs the same task as CreateElement but calls a
+// continuation function after the child element is created, allowing
+// additional actions to be performed on the child element before returning.
+//
+// This method of element creation is particularly useful when building nested
+// XML documents from code. For example:
+//
+//	org := doc.CreateChild("organization", func(e *Element) {
+//		e.CreateComment("Mary")
+//		e.CreateChild("person", func(e *Element) {
+//			e.CreateAttr("name", "Mary")
+//			e.CreateAttr("age", "30")
+//			e.CreateAttr("hair", "brown")
+//		})
+//	})
+func (e *Element) CreateChild(tag string, cont func(e *Element)) *Element {
+	child := e.CreateElement(tag)
+	cont(child)
+	return child
+}
+
 // AddChild adds the token 't' as the last child of the element. If token 't'
 // was already the child of another element, it is first removed from its
 // parent element.
@@ -751,7 +831,7 @@ func (e *Element) InsertChildAt(index int, t Token) {
 	}
 
 	if t.Parent() != nil {
-		if t.Parent() == e && t.Index() > index {
+		if t.Parent() == e && t.Index() < index {
 			index--
 		}
 		t.Parent().RemoveChild(t)
@@ -796,6 +876,27 @@ func (e *Element) RemoveChildAt(index int) Token {
 	return t
 }
 
+// autoClose analyzes the stack's top element and the current token to decide
+// whether the top element should be closed.
+func (e *Element) autoClose(stack *stack[*Element], t xml.Token, tags []string) {
+	if stack.empty() {
+		return
+	}
+
+	top := stack.peek()
+
+	for _, tag := range tags {
+		if strings.EqualFold(tag, top.FullTag()) {
+			if e, ok := t.(xml.EndElement); !ok ||
+				!strings.EqualFold(e.Name.Space, top.Space) ||
+				!strings.EqualFold(e.Name.Local, top.Tag) {
+				stack.pop()
+			}
+			break
+		}
+	}
+}
+
 // ReadFrom reads XML from the reader 'ri' and stores the result as a new
 // child of this element.
 func (e *Element) readFrom(ri io.Reader, settings ReadSettings) (n int64, err error) {
@@ -808,12 +909,10 @@ func (e *Element) readFrom(ri io.Reader, settings ReadSettings) (n int64, err er
 		r = newXmlSimpleReader(ri)
 	}
 
-	dec := xml.NewDecoder(r)
-	dec.CharsetReader = settings.CharsetReader
-	dec.Strict = !settings.Permissive
-	dec.Entity = settings.Entity
+	attrCheck := make(map[xml.Name]int)
+	dec := newDecoder(r, settings)
 
-	var stack stack
+	var stack stack[*Element]
 	stack.push(e)
 	for {
 		if pr != nil {
@@ -822,6 +921,10 @@ func (e *Element) readFrom(ri io.Reader, settings ReadSettings) (n int64, err er
 
 		t, err := dec.RawToken()
 
+		if settings.Permissive && settings.AutoClose != nil {
+			e.autoClose(&stack, t, settings.AutoClose)
+		}
+
 		switch {
 		case err == io.EOF:
 			if len(stack.data) != 1 {
@@ -834,13 +937,24 @@ func (e *Element) readFrom(ri io.Reader, settings ReadSettings) (n int64, err er
 			return r.Bytes(), ErrXML
 		}
 
-		top := stack.peek().(*Element)
+		top := stack.peek()
 
 		switch t := t.(type) {
 		case xml.StartElement:
 			e := newElement(t.Name.Space, t.Name.Local, top)
-			for _, a := range t.Attr {
-				e.createAttr(a.Name.Space, a.Name.Local, a.Value, e, settings.PreserveDuplicateAttrs)
+			if settings.PreserveDuplicateAttrs || len(t.Attr) < 2 {
+				for _, a := range t.Attr {
+					e.addAttr(a.Name.Space, a.Name.Local, a.Value)
+				}
+			} else {
+				for _, a := range t.Attr {
+					if i, contains := attrCheck[a.Name]; contains {
+						e.Attr[i].Value = a.Value
+					} else {
+						attrCheck[a.Name] = e.addAttr(a.Name.Space, a.Name.Local, a.Value)
+					}
+				}
+				clear(attrCheck)
 			}
 			stack.push(e)
 		case xml.EndElement:
@@ -970,6 +1084,25 @@ func (e *Element) FindElementsPath(path Path) []*Element {
 	return p.traverse(e, path)
 }
 
+// NotNil returns the receiver element if it isn't nil; otherwise, it returns
+// an unparented element with an empty string tag. This function simplifies
+// the task of writing code to ignore not-found results from element queries.
+// For example, instead of writing this:
+//
+//	if e := doc.SelectElement("enabled"); e != nil {
+//		e.SetText("true")
+//	}
+//
+// You could write this:
+//
+//	doc.SelectElement("enabled").NotNil().SetText("true")
+func (e *Element) NotNil() *Element {
+	if e == nil {
+		return NewElement("")
+	}
+	return e
+}
+
 // GetPath returns the absolute path of the element. The absolute path is the
 // full path from the document's root.
 func (e *Element) GetPath() string {
@@ -1173,6 +1306,34 @@ func (e *Element) dup(parent *Element) Token {
 	return ne
 }
 
+// NextSibling returns this element's next sibling element. It returns nil if
+// there is no next sibling element.
+func (e *Element) NextSibling() *Element {
+	if e.parent == nil {
+		return nil
+	}
+	for i := e.index + 1; i < len(e.parent.Child); i++ {
+		if s, ok := e.parent.Child[i].(*Element); ok {
+			return s
+		}
+	}
+	return nil
+}
+
+// PrevSibling returns this element's preceding sibling element. It returns
+// nil if there is no preceding sibling element.
+func (e *Element) PrevSibling() *Element {
+	if e.parent == nil {
+		return nil
+	}
+	for i := e.index - 1; i >= 0; i-- {
+		if s, ok := e.parent.Child[i].(*Element); ok {
+			return s
+		}
+	}
+	return nil
+}
+
 // Parent returns this element's parent element. It returns nil if this
 // element has no parent.
 func (e *Element) Parent() *Element {
@@ -1236,28 +1397,29 @@ func (e *Element) addChild(t Token) {
 // prefix followed by a colon.
 func (e *Element) CreateAttr(key, value string) *Attr {
 	space, skey := spaceDecompose(key)
-	return e.createAttr(space, skey, value, e, false)
-}
 
-// createAttr is a helper function that creates attributes.
-func (e *Element) createAttr(space, key, value string, parent *Element, preserveDups bool) *Attr {
-	if !preserveDups {
-		for i, a := range e.Attr {
-			if space == a.Space && key == a.Key {
-				e.Attr[i].Value = value
-				return &e.Attr[i]
-			}
+	for i, a := range e.Attr {
+		if space == a.Space && skey == a.Key {
+			e.Attr[i].Value = value
+			return &e.Attr[i]
 		}
 	}
 
+	i := e.addAttr(space, skey, value)
+	return &e.Attr[i]
+}
+
+// addAttr is a helper function that adds an attribute to an element. Returns
+// the index of the added attribute.
+func (e *Element) addAttr(space, key, value string) int {
 	a := Attr{
 		Space:   space,
 		Key:     key,
 		Value:   value,
-		element: parent,
+		element: e,
 	}
 	e.Attr = append(e.Attr, a)
-	return &e.Attr[len(e.Attr)-1]
+	return len(e.Attr) - 1
 }
 
 // RemoveAttr removes the first attribute of this element whose key matches
@@ -1282,25 +1444,12 @@ func (e *Element) RemoveAttr(key string) *Attr {
 
 // SortAttrs sorts this element's attributes lexicographically by key.
 func (e *Element) SortAttrs() {
-	sort.Sort(byAttr(e.Attr))
-}
-
-type byAttr []Attr
-
-func (a byAttr) Len() int {
-	return len(a)
-}
-
-func (a byAttr) Swap(i, j int) {
-	a[i], a[j] = a[j], a[i]
-}
-
-func (a byAttr) Less(i, j int) bool {
-	sp := strings.Compare(a[i].Space, a[j].Space)
-	if sp == 0 {
-		return strings.Compare(a[i].Key, a[j].Key) < 0
-	}
-	return sp < 0
+	slices.SortFunc(e.Attr, func(a, b Attr) int {
+		if v := strings.Compare(a.Space, b.Space); v != 0 {
+			return v
+		}
+		return strings.Compare(a.Key, b.Key)
+	})
 }
 
 // FullKey returns this attribute's complete key, including namespace prefix
@@ -1336,7 +1485,7 @@ func (a *Attr) WriteTo(w Writer, s *WriteSettings) {
 		w.WriteString(`="`)
 	}
 	var m escapeMode
-	if s.CanonicalAttrVal {
+	if s.CanonicalAttrVal && !s.AttrSingleQuote {
 		m = escapeCanonicalAttr
 	} else {
 		m = escapeNormal
